#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle
import weakref

import numpy as np
cimport numpy as np
cimport cython
try:
    import pandas as pd
except ImportError:  # pragma: no cover
    pd = None

from cpython.version cimport PY_MAJOR_VERSION

from ..compat import six, OrderedDict
from ..core import BaseWithKey
from ..utils_c cimport to_str
from .core cimport ProviderType, ValueType, Identity, List, Tuple, Dict, \
    Reference, KeyPlaceholder, AttrWrapper, Provider, Field, \
    OneOfField, ReferenceField, IdentityField, \
    ListField, TupleField
from .protos.value_pb2 import Value
from .dataserializer import dumps as datadumps, loads as dataloads


cdef dict PRIMITIVE_TYPE_TO_VALUE_FIELD = {
    ValueType.bool: 'b',
    ValueType.int8: 'i',
    ValueType.int16: 'i',
    ValueType.int32: 'i',
    ValueType.int64: 'i',
    ValueType.uint8: 'i',
    ValueType.uint16: 'i',
    ValueType.uint32: 'i',
    ValueType.uint64: 'i',
    ValueType.float16: 'f',
    ValueType.float32: 'f',
    ValueType.float64: 'f',
    ValueType.bytes: 's',
    ValueType.unicode: 'u',
}


cdef class ProtobufSerializeProvider(Provider):
    def __init__(self):
        self.type = ProviderType.protobuf

    cdef inline void _set_slice(self, slice value, obj, tp=None):
        if value.start is not None:
            obj.slice.start_val = value.start
        if value.stop is not None:
            obj.slice.stop_val = value.stop
        if value.step is not None:
            obj.slice.step_val = value.step
        if all(it is None for it in (value.start, value.stop, value.step)):
            obj.slice.is_null = True

    cdef inline slice _get_slice(self, obj):
        if obj.slice.is_null:
            return slice(None)

        start = obj.slice.start_val if obj.slice.WhichOneof('start') else None
        stop = obj.slice.stop_val if obj.slice.WhichOneof('stop') else None
        step = obj.slice.step_val if obj.slice.WhichOneof('step') else None
        return slice(start, stop, step)

    cdef inline void _set_arr(self, np.ndarray value, obj, tp=None):
        obj.arr = datadumps(value)

    cdef inline np.ndarray _get_arr(self, obj):
        cdef object x

        x = obj.arr
        return dataloads(x) if x is not None and len(x) > 0 else None

    cdef inline void _set_dtype(self, np.dtype value, obj, tp=None):
        if 'V' not in value.str:
            dtype = value.str
            if isinstance(dtype, unicode):
                dtype = dtype.encode('utf-8')
            obj.dtype = dtype
        else:
            obj.dtype = pickle.dumps(value)

    cdef inline np.dtype _get_dtype(self, obj):
        if obj.dtype is None or len(obj.dtype) == 0:
            return
        try:
            return np.dtype(obj.dtype)
        except TypeError:
            return np.dtype(pickle.loads(obj.dtype))

    cdef inline void _set_index(self, value, obj, tp=None) except *:
        obj.index = datadumps(value)

    cdef inline _get_index(self, obj):
        x = obj.index
        return dataloads(x) if x is not None and len(x) > 0 else None

    cdef inline void _set_series(self, value, obj, tp=None) except *:
        obj.series = datadumps(value)

    cdef inline _get_series(self, obj):
        x = obj.series
        return dataloads(x) if x is not None and len(x) > 0 else None

    cdef inline void _set_dataframe(self, value, obj, tp=None) except *:
        obj.dataframe = datadumps(value)

    cdef inline object _get_dataframe(self, obj):
        x = obj.dataframe
        return dataloads(x) if x is not None and len(x) > 0 else None

    cdef inline void _set_key(self, object value, obj, tp=None) except *:
        obj.key.key = value.key
        obj.key.id = value.id

    cdef inline KeyPlaceholder _get_key(self, obj):
        if obj.key.key is None or len(obj.key.key) == 0:
            return
        if PY_MAJOR_VERSION >= 3:
            return KeyPlaceholder(obj.key.key, obj.key.id)
        else:
            return KeyPlaceholder(to_str(obj.key.key), to_str(obj.key.id))

    cdef inline void _set_datetime64(self, value, obj, tp=None):
        bio = six.BytesIO()
        np.save(bio, value)
        obj.datetime64 = bio.getvalue()

    cdef inline object _get_datetime64(self, obj):
        cdef object x

        x = obj.datetime64
        return np.load(six.BytesIO(x)) if x is not None and len(x) > 0 else None

    cdef inline void _set_timedelta64(self, value, obj, tp=None):
        bio = six.BytesIO()
        np.save(bio, value)
        obj.timedelta64 = bio.getvalue()

    cdef inline object _get_timedelta64(self, obj):
        cdef object x

        x = obj.timedelta64
        return np.load(six.BytesIO(x)) if x is not None and len(x) > 0 else None

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    cdef inline void _set_list(self, list value, obj, tp=None, bint weak_ref=False) except *:
        cdef list res

        # list is special, cuz the internal object can be a specified type or any type
        if isinstance(obj, Value):
            for val in value:
                if weak_ref:
                    val = val()
                it_obj = obj.list.value.add()
                self._set_value(val, it_obj, tp=tp.type if tp is not None else tp)
        else:
            try:
                obj.extend(value)
            except TypeError:
                # not primitive type
                res = []
                for val in value:
                    if weak_ref:
                        val = val()
                    it_obj = Value()
                    self._set_value(val, it_obj, tp=tp.type if tp is not None else tp)
                    res.append(it_obj)
                obj.extend(res)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    cdef inline list _get_list(self, obj, tp, list callbacks, bint weak_ref):
        cdef list res
        cdef int i
        cdef object it_obj

        obj = obj.list.value if isinstance(obj, Value) else obj
        res = []
        for i, it_obj in enumerate(obj):
            if not isinstance(it_obj, Value):
                res.append(it_obj if not weak_ref else weakref.ref(it_obj))
                continue

            val = self._get_value(it_obj, tp.type if tp is not None else tp,
                                  callbacks, weak_ref)
            res.append(val)
            if isinstance(val, KeyPlaceholder):
                def cb(j, v):
                    def inner(subs):
                        o = subs[v.key, v.id]
                        if weak_ref:
                            o = weakref.ref(o)
                        res[j] = o
                    return inner
                callbacks.append(cb(i, val))
        return res

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    cdef inline void _set_tuple(self, tuple value, obj, tp=None, bint weak_ref=False) except *:
        cdef list res

        if isinstance(obj, Value):
            obj.list.is_tuple = True
            if tp is not None and isinstance(tp.type, tuple):
                if len(tp.type) != len(value):
                    raise ValueError('Value length should be {0} '
                                     'according to type definition, got {1}'
                                     .format(len(tp.type), len(value)))
                for it_type, val in zip(tp.type, value):
                    if weak_ref:
                        val = val()
                    if it_type == ValueType.key:
                        raise TypeError('`key` type is not allowed in a tuple')
                    it_obj = obj.list.value.add()
                    self._set_typed_value(val, it_obj, it_type)
            else:
                if tp == ValueType.key:
                    raise TypeError('`key` type is not allowed in a tuple')
                for val in value:
                    if weak_ref:
                        val = val()
                    it_obj = obj.list.value.add()
                    self._set_value(val, it_obj, tp=tp.type if tp is not None else tp)
        else:
            try:
                if weak_ref:
                    value = tuple(val() for val in value)
                obj.extend(value)
            except TypeError:
                # not primitive type
                res = []
                for val in value:
                    if weak_ref:
                        val = val()
                    it_obj = Value()
                    self._set_value(val, it_obj, tp=tp.type if tp is not None else tp)
                    res.append(it_obj)
                obj.extend(res)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    cdef inline tuple _get_tuple(self, obj, tp, list callbacks, bint weak_ref):
        cdef list res

        values = obj.list.value if isinstance(obj, Value) else obj

        res = []
        if tp is not None and isinstance(tp.type, tuple):
            for it_type, it_obj in zip(tp.type, values):
                if isinstance(it_obj, Value):
                    res.append(self._get_value(it_obj, it_type,
                                               callbacks, weak_ref))
                else:
                    if weak_ref:
                        it_obj = weakref.ref(it_obj)
                    res.append(it_obj)
        else:
            for it_obj in values:
                if isinstance(it_obj, Value):
                    res.append(self._get_value(it_obj, tp.type if tp is not None else tp,
                                               callbacks, weak_ref))
                else:
                    if weak_ref:
                        it_obj = weakref.ref(it_obj)
                    res.append(it_obj)

        return tuple(res)

    cdef inline void _set_dict(self, value, obj, tp=None, bint weak_ref=False) except *:
        for k, v in value.items():
            if weak_ref:
                k, v = k(), v()
            key_obj = obj.dict.keys.value.add()
            if tp is not None and tp.key_type == ValueType.key:
                raise TypeError('`key` type is not allowd as dict key')
            self._set_value(k, key_obj, tp=tp.key_type if tp is not None else tp)
            value_obj = obj.dict.values.value.add()
            self._set_value(v, value_obj, tp=tp.value_type if tp is not None else tp)

    cdef inline object _get_dict(self, obj, tp, list callbacks, bint weak_ref):
        d = OrderedDict()
        for k_obj, v_obj in zip(obj.dict.keys.value, obj.dict.values.value):
            k = self._get_value(k_obj, tp.key_type if tp is not None else tp,
                                callbacks, weak_ref)
            v = self._get_value(v_obj, tp.value_type if tp is not None else tp,
                                callbacks, weak_ref)
            d[k] = v
            if isinstance(v, KeyPlaceholder):
                def cb(key, value):
                    def inner(subs):
                        o = subs[value.key, value.id]
                        if weak_ref:
                            o = weakref.ref(o)
                        d[key] = o
                    return inner
                callbacks.append(cb(k, v))

        return d

    cdef inline void _set_typed_value(self, value, obj, tp, bint weak_ref=False) except *:
        cdef str value_field

        if type(tp) not in (List, Tuple, Dict) and weak_ref:
            # not iterable, and is weak ref
            value = value()

        if value is None:
            # handle None
            obj.is_null = True
            return

        if tp in PRIMITIVE_TYPE_TO_VALUE_FIELD:
            value_field = PRIMITIVE_TYPE_TO_VALUE_FIELD[tp]
            setattr(obj, value_field, value)
        elif tp is ValueType.slice:
            self._set_slice(value, obj, tp)
        elif tp is ValueType.arr:
            self._set_arr(<np.ndarray>value, obj, tp)
        elif tp is ValueType.dtype:
            self._set_dtype(<np.dtype>value, obj, tp)
        elif tp is ValueType.index:
            self._set_index(value, obj, tp)
        elif tp is ValueType.series:
            self._set_series(value, obj, tp)
        elif tp is ValueType.dataframe:
            self._set_dataframe(value, obj, tp)
        elif tp is ValueType.key:
            self._set_key(value, obj, tp)
        elif tp is ValueType.datetime64:
            self._set_datetime64(value, obj, tp)
        elif tp is ValueType.timedelta64:
            self._set_timedelta64(value, obj, tp)
        elif isinstance(tp, Identity):
            value_field = PRIMITIVE_TYPE_TO_VALUE_FIELD[tp.type]
            setattr(obj, value_field, value)
        elif isinstance(tp, List):
            # list type
            self._set_list(<list>value, obj, tp, weak_ref=weak_ref)
        elif isinstance(tp, Tuple):
            # tuple type
            self._set_tuple(<tuple>value, obj, tp, weak_ref=weak_ref)
        elif isinstance(tp, Dict):
            # dict type
            self._set_dict(<dict>value, obj, tp, weak_ref=weak_ref)
        else:
            raise TypeError('Unknown type to serialize: {0}'.format(tp))

    cdef inline void _set_untyped_value(self, value, obj, bint weak_ref=False) except *:
        if not isinstance(value, (list, tuple, dict)) and weak_ref:
            # not iterable, and is weak ref
            value = value()

        # we are not aware of the type, so try to infer it
        if value is None:
            # handle None
            obj.is_null = True
            return

        if isinstance(value, bool):
            obj.b = value
        elif isinstance(value, bytes):
            obj.s = value
        elif isinstance(value, unicode):
            obj.u = value
        elif isinstance(value, int):
            obj.i = value
        elif isinstance(value, float):
            obj.f = value
        elif isinstance(value, slice):
            self._set_slice(value, obj)
        elif isinstance(value, np.ndarray):
            self._set_arr(value, obj)
        elif isinstance(value, np.dtype):
            self._set_dtype(value, obj)
        elif pd is not None and isinstance(value, pd.Index):
            self._set_index(value, obj)
        elif pd is not None and isinstance(value, pd.Series):
            self._set_series(value, obj)
        elif pd is not None and isinstance(value, pd.DataFrame):
            self._set_dataframe(value, obj)
        elif isinstance(value, BaseWithKey):
            self._set_key(value, obj)
        elif isinstance(value, list):
            self._set_list(value, obj, tp=None, weak_ref=weak_ref)
        elif isinstance(value, tuple):
            self._set_tuple(value, obj, tp=None, weak_ref=weak_ref)
        elif isinstance(value, dict):
            self._set_dict(value, obj, tp=None, weak_ref=weak_ref)
        elif isinstance(value, np.datetime64):
            self._set_datetime64(value, obj)
        elif isinstance(value, np.timedelta64):
            self._set_timedelta64(value, obj)
        elif isinstance(value, np.number):
            self._set_untyped_value(value.item(), obj)
        elif PY_MAJOR_VERSION < 3 and isinstance(value, long):
            obj.i = value
        else:
            raise TypeError('Unknown type to serialize: {0}'.format(type(value)))

    cdef inline void _set_value(cls, value, obj, tp=None, bint weak_ref=False) except *:
        if tp is None:
            cls._set_untyped_value(value, obj, weak_ref=weak_ref)
        else:
            cls._set_typed_value(value, obj, tp, weak_ref=weak_ref)

    cpdef serialize_field(self, Field field, model_instance, obj):
        cdef object value
        cdef object val
        cdef Field f
        cdef str tag
        cdef object field_obj
        cdef object add
        cdef object it_obj
        cdef OneOfField oneoffield

        if isinstance(field, OneOfField):
            oneoffield = <OneOfField> field
            value = getattr(model_instance, oneoffield.attr, None)
            for f in oneoffield.fields:
                if isinstance(value, f.type.model):
                    f.serialize(self, model_instance, obj)
                    return
            return

        value = getattr(model_instance, field.attr, field.default)
        if field.on_serialize and value is not None:
            value = field.on_serialize(value)
        tag = field.tag_name(self)
        field_obj = getattr(obj, tag)

        if value is None:
            return

        if isinstance(field, ReferenceField):
            if field.weak_ref:
                field_obj = field_obj()
            value.serialize(self, obj=field_obj)
        elif isinstance(field, ListField):
            if type(field.type.type) == Reference:
                add = field_obj.list.value.add if isinstance(field_obj, Value) else field_obj.add
                for val in value:
                    if field.weak_ref:
                        val = val()
                    it_obj = add()
                    if val is not None:
                        val.serialize(self, obj=it_obj)
                    elif isinstance(it_obj, Value):
                        it_obj.is_null = True
            else:
                # repeated primitive type
                if not isinstance(value, list):
                    value = list(value)
                self._set_list(value, field_obj, tp=field.type, weak_ref=field.weak_ref)
        elif isinstance(field_obj, Value):
            self._set_value(value, field_obj, field.type, weak_ref=field.weak_ref)
        elif isinstance(field, TupleField):
            self._set_tuple(value, field_obj, tp=field.type, weak_ref=field.weak_ref)
        else:
            setattr(obj, tag, value)

    cpdef serialize_attribute_as_dict(self, model_instance, obj=None):
        cdef object id_field
        cdef str attr
        cdef object d_obj
        cdef dict fields
        cdef str name
        cdef object field
        cdef object value

        if obj is None:
            obj = model_instance.cls(self)()

        if hasattr(model_instance, '_ID_FIELD') and not isinstance(obj, Value):
            for id_field in model_instance._ID_FIELD:
                id_field.serialize(self, model_instance, obj)

        attr = getattr(model_instance, 'attr_tag', None)
        if attr:
            d_obj = getattr(obj, attr)
        else:
            d_obj = obj

        if isinstance(d_obj, Value):
            fields = model_instance._FIELDS
            for name, field in fields.items():
                value = getattr(model_instance, name, None)
                if value is None:
                    continue

                k = d_obj.dict.keys.value.add()
                self._set_value(field.tag_name(self), k, tp=ValueType.string)
                v = d_obj.dict.values.value.add()
                self._set_value(value, v, tp=field.type, weak_ref=field.weak_ref)
        else:
            fields = model_instance._FIELDS
            for name, field in fields.items():
                value = getattr(model_instance, name, None)
                if value is not None:
                    field.serialize(self, model_instance, AttrWrapper(d_obj))

        return obj

    cdef inline object _get_typed_value(self, obj, tp, list callbacks, bint weak_ref):
        cdef str value_field

        if obj.is_null:
            return

        if weak_ref:
            ref = weakref.ref
        else:
            ref = lambda x: x

        if tp in PRIMITIVE_TYPE_TO_VALUE_FIELD:
            value_field = PRIMITIVE_TYPE_TO_VALUE_FIELD[tp]
            return ref(getattr(obj, value_field))
        elif tp is ValueType.slice:
            return ref(self._get_slice(obj))
        elif tp is ValueType.arr:
            return ref(self._get_arr(obj))
        elif tp is ValueType.dtype:
            return ref(self._get_dtype(obj))
        elif tp is ValueType.index:
            return ref(self._get_index(obj))
        elif tp is ValueType.series:
            return ref(self._get_series(obj))
        elif tp is ValueType.dataframe:
            return ref(self._get_dataframe(obj))
        elif tp is ValueType.key:
            return self._get_key(obj)
        elif tp is ValueType.datetime64:
            return ref(self._get_datetime64(obj))
        elif tp is ValueType.timedelta64:
            return ref(self._get_timedelta64(obj))
        elif isinstance(tp, Identity):
            value_field = PRIMITIVE_TYPE_TO_VALUE_FIELD[tp.type]
            return ref(getattr(obj, value_field))
        elif isinstance(tp, List):
            # list type
            return self._get_list(obj, tp, callbacks, weak_ref)
        elif isinstance(tp, Tuple):
            # tuple type
            return self._get_tuple(obj, tp, callbacks, weak_ref)
        elif isinstance(tp, Dict):
            # dict type
            return self._get_dict(obj, tp, callbacks, weak_ref)
        else:
            raise TypeError('Unknown type to deserialize: {0}'.format(tp))

    cdef inline object _get_untyped_value(self, obj, list callbacks, bint weak_ref):
        cdef str field

        if weak_ref:
            ref = weakref.ref
        else:
            ref = lambda x: x

        # we are not aware of the type, so try to infer it
        field = obj.WhichOneof('value')

        if field is None or field == 'is_null':
            return
        elif field in 'bif':
            # primitive type
            return ref(getattr(obj, field))
        elif field == 's':
            # bytes
            b = obj.s
            if not isinstance(b, bytes):
                b = b.encode('utf-8')
            return ref(b)
        elif field == 'u':
            # unicode
            u = obj.u
            if not isinstance(u, unicode):
                u = u.decode('utf-8')
            return ref(u)
        elif field == 'list':
            if obj.list.is_tuple:
                return self._get_tuple(obj, None, callbacks, weak_ref)
            else:
                return self._get_list(obj, None, callbacks, weak_ref)
        elif field == 'dict':
            return self._get_dict(obj, None, callbacks, weak_ref)
        elif field == 'slice':
            return ref(self._get_slice(obj))
        elif field == 'arr':
            return ref(self._get_arr(obj))
        elif field == 'dtype':
            return ref(self._get_dtype(obj))
        elif field == 'index':
            return ref(self._get_index(obj))
        elif field == 'series':
            return ref(self._get_series(obj))
        elif field == 'dataframe':
            return ref(self._get_dataframe(obj))
        elif field == 'key':
            return self._get_key(obj)
        elif field == 'datetime64':
            return ref(self._get_datetime64(obj))
        elif field == 'timedelta64':
            return ref(self._get_timedelta64(obj))
        else:
            raise TypeError('Unknown type to deserialize')

    cdef inline object _get_value(self, obj, tp, list callbacks, bint weak_ref):
        if tp is None:
            return self._get_untyped_value(obj, callbacks, weak_ref)
        else:
            return self._get_typed_value(obj, tp, callbacks, weak_ref)

    cdef inline object _on_deserial(self, Field field, x):
        x = x if x is not None else field.default
        if field.on_deserialize is None:
            return x
        return field.on_deserialize(x)

    def deserialize_field(self, Field field, model_instance, obj, list callbacks, dict key_to_instance):
        cdef str tag
        cdef str f_tag
        cdef Field f
        cdef object f_obj
        cdef object field_obj
        cdef OneOfField oneoffield

        if isinstance(field, OneOfField):
            oneoffield = <OneOfField> field
            tag = oneoffield.tag_name(self)
            f_tag = obj.WhichOneof(tag)
            if f_tag is None:
                return
            f = next(f for f in oneoffield.fields if f.tag_name(self) == f_tag)
            f_obj = getattr(obj, f_tag)
            setattr(model_instance, f.attr,
                    f.type.model.deserialize(self, f_obj, callbacks, key_to_instance))
            return

        tag = field.tag_name(self)
        field_obj = getattr(obj, tag)

        if isinstance(field, ReferenceField):
            setattr(model_instance, field.attr,
                    self._on_deserial(
                        field, field.type.model.deserialize(self, field_obj, callbacks, key_to_instance)))
        elif isinstance(field, ListField):
            if type(field.type.type) == Reference:
                field_obj = field_obj.list.value if isinstance(field_obj, Value) else field_obj
                setattr(model_instance, field.attr,
                        self._on_deserial(
                            field, [field.type.type.model.deserialize(self, it_obj, callbacks, key_to_instance)
                                    for it_obj in field_obj]))
            else:
                setattr(model_instance, field.attr,
                        self._get_list(field_obj, field.type, callbacks, field.weak_ref))
        elif isinstance(field_obj, Value):
            val = self._get_value(field_obj, field.type, callbacks, field.weak_ref)
            setattr(model_instance, field.attr, self._on_deserial(field, val))
            if isinstance(val, KeyPlaceholder):
                def cb(subs):
                    o = self._on_deserial(field, subs[val.key, val.id])
                    if field.weak_ref:
                        o = weakref.ref(o)
                    setattr(model_instance, field.attr, o)

                callbacks.append(cb)
        elif isinstance(field, TupleField):
            setattr(model_instance, field.attr,
                    self._on_deserial(field, self._get_tuple(field_obj, field.type, callbacks,
                                                             field.weak_ref)))
        else:
            setattr(model_instance, field.attr, self._on_deserial(field, getattr(obj, tag)))

    def deserialize_attribute_as_dict(self, model_cls, obj, list callbacks, dict key_to_instance):
        cdef str attr
        cdef object d_obj
        cdef object kw
        cdef tuple id_fields
        cdef IdentityField id_field
        cdef object model_instance
        cdef dict tag_to_fields
        cdef Field field
        cdef Field f
        cdef OneOfField oneoffield
        cdef Field it_field
        cdef str tag
        cdef object o_tag

        attr = getattr(model_cls, 'attr_tag', None)
        if attr:
            d_obj = getattr(obj, attr)
        else:
            d_obj = obj

        kw = AttrWrapper(dict())
        if hasattr(model_cls, '_ID_FIELD'):
            # get the id field
            if attr is None:
                if d_obj.is_null:
                    return
                id_fields = model_cls._ID_FIELD
                for id_field in id_fields:
                    tag = id_field.tag_name(self)
                    v = next(v for k, v in zip(d_obj.dict.keys.value, d_obj.dict.values.value)
                             if self._get_value(k, ValueType.string, callbacks, id_field.weak_ref) == tag)
                    setattr(kw, id_field.attr, self._get_value(v, id_field.type,
                                                               callbacks, id_field.weak_ref))
            else:
                for id_field in model_cls._ID_FIELD:
                    id_field.deserialize(self, kw, obj, callbacks, key_to_instance)
        model_instance = model_cls(**kw.asdict())

        tag_to_fields = dict()
        for field in six.itervalues(model_instance._FIELDS):
            if isinstance(field, OneOfField):
                oneoffield = <OneOfField> field
                for f in oneoffield.fields:
                    tag_to_fields[f.tag_name(self)] = f
            else:
                tag_to_fields[field.tag_name(self)] = field

        if isinstance(d_obj, Value):
            if d_obj.is_null:
                return
            for k, v in zip(d_obj.dict.keys.value, d_obj.dict.values.value):
                tag = self._get_value(k, ValueType.string, callbacks, False)
                it_field = tag_to_fields[tag]
                setattr(model_instance, it_field.attr,
                        self._get_value(v, it_field.type, callbacks, it_field.weak_ref))
        else:
            for o_tag in d_obj:
                it_field = tag_to_fields[o_tag]
                it_field.deserialize(self, model_instance, AttrWrapper(d_obj),
                                     callbacks, key_to_instance)

        if isinstance(model_instance, BaseWithKey):
            key_to_instance[model_instance.key, model_instance.id] = model_instance
        return model_instance



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import logging
import os
import time

from .kvstore import KVStoreActor
from .utils import SchedulerActor
from ..config import options


logger = logging.getLogger(__name__)


class ResourceActor(SchedulerActor):
    """
    Actor managing free resources on workers
    """
    def __init__(self):
        super(ResourceActor, self).__init__()
        self._meta_cache = dict()
        self._kv_store_ref = None

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        super(ResourceActor, self).post_create()
        self.ref().clean_worker()

        self._kv_store_ref = self.ctx.actor_ref(KVStoreActor.default_name())
        if not self.ctx.has_actor(self._kv_store_ref):
            self._kv_store_ref = None

    @classmethod
    def default_name(cls):
        return 's:%s' % cls.__name__

    def clean_worker(self):
        """
        Remove worker when it does not update its status for a long time
        """
        for worker in list(self._meta_cache.keys()):
            worker_meta = self._meta_cache[worker]
            if 'update_time' not in worker_meta:
                continue

            if time.time() - worker_meta['update_time'] > options.scheduler.status_timeout:
                del self._meta_cache[worker]

        self.ref().clean_worker(_tell=True, _delay=1)

    def get_worker_count(self):
        return len(self._meta_cache)

    def get_workers_meta(self):
        return copy.deepcopy(self._meta_cache)

    def set_worker_meta(self, worker, worker_meta):
        self._meta_cache[worker] = worker_meta
        if self._kv_store_ref is not None:
            self._kv_store_ref.write('/workers/meta/%s' % worker, json.dumps(worker_meta),
                                     _tell=True, _wait=False)
            self._kv_store_ref.write('/workers/meta_timestamp', str(int(time.time())),
                                     _tell=True, _wait=False)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import heapq
import logging
import time

from .utils import WorkerActor
from .. import resource
from ..utils import log_unhandled

logger = logging.getLogger(__name__)
_ALLOCATE_PERIOD = 0.5


class ChunkPriorityItem(object):
    """
    Class providing an order for operands for assignment
    """
    __slots__ = '_op_key', '_session_id', '_priority', '_callback'

    def __init__(self, session_id, op_key, priority_data, callback):
        self._op_key = op_key
        self._session_id = session_id
        self._callback = callback

        self._priority = ()
        self.update_priority(priority_data)

    def update_priority(self, priority_data, copy=False):
        """
        Update priority data in the item
        :param priority_data: priority data
        :param copy: if True, the function will return a new item, otherwise the update will be applied locally
        """
        if copy:
            return ChunkPriorityItem(self._session_id, self._op_key, priority_data,
                                     self._callback)
        else:
            self._priority = tuple([
                priority_data.get('depth', 0),
                priority_data.get('demand_depths', ()),
                -priority_data.get('successor_size', 0),
                -priority_data.get('placement_order', 0),
                priority_data.get('descendant_size'),
            ])
            return self

    @property
    def session_id(self):
        return self._session_id

    @property
    def op_key(self):
        return self._op_key

    @property
    def callback(self):
        return self._callback

    def __lt__(self, other):
        """
        :type other: ChunkPriorityItem
        """
        return self._priority > other._priority


class TaskQueueActor(WorkerActor):
    """
    Actor accepting requests and holding the queue
    """
    def __init__(self, parallel_num=None):
        super(TaskQueueActor, self).__init__()
        self._requests = dict()
        self._allocated = set()
        self._allocate_pendings = set()
        self._req_heap = []

        self._allocator_ref = None
        self._parallel_num = parallel_num or resource.cpu_count()

    def post_create(self):
        self._allocator_ref = self.ctx.create_actor(
            TaskQueueAllocatorActor, self.ref(), self._parallel_num,
            uid=TaskQueueAllocatorActor.default_name())

    def enqueue_task(self, session_id, op_key, priority_data, callback):
        """
        Put a task in queue for allocation
        :param session_id: session id
        :param op_key: operand key
        :param priority_data: priority data
        :param callback: callback to invoke when the resources are allocated
        """
        logger.debug('Operand task %s enqueued.', op_key)
        item = ChunkPriorityItem(session_id, op_key, priority_data, callback)
        self._requests[(session_id, op_key)] = item
        heapq.heappush(self._req_heap, item)

        self._allocator_ref.allocate_tasks(_tell=True)

    def update_priority(self, session_id, op_key, priority_data):
        """
        Update priority data for the specified operand
        :param session_id: session id
        :param op_key: operand key
        :param priority_data: new priority data
        """
        logger.debug('Priority data for operand task %s updated.', op_key)
        query_key = (session_id, op_key)
        if query_key not in self._requests:
            return
        item = self._requests[query_key]
        item = item.update_priority(priority_data, copy=True)
        self._requests[query_key] = item
        heapq.heappush(self._req_heap, item)

    def mark_allocate_pending(self, session_id, op_key):
        """
        Mark an operand as being allocated, i.e., it has been submitted to the MemQuotaActor.
        :param session_id: session id
        :param op_key: operand key
        """
        self._allocate_pendings.add((session_id, op_key))

    def handle_allocated(self, session_id, op_key, callback, *args, **kwargs):
        """
        When MemQuotaActor allocates resource for an operand, put the operand into
        allocated and then invoke the callback
        :param session_id: session id
        :param op_key: operand key
        :param callback: callback to invoke
        """
        logger.debug('Operand task %s allocated.', op_key)
        query_key = (session_id, op_key)
        self.tell_promise(callback, *args, **kwargs)
        try:
            self._allocate_pendings.remove(query_key)
        except KeyError:
            pass
        self._allocated.add(query_key)

    def release_task(self, session_id, op_key):
        """
        Remove an operand task from queue
        :param session_id: session id
        :param op_key: operand key
        """
        logger.debug('Operand task %s released.', op_key)
        query_key = (session_id, op_key)
        try:
            del self._requests[(session_id, op_key)]
        except KeyError:
            pass
        try:
            self._allocated.remove(query_key)
        except KeyError:
            pass
        try:
            self._allocate_pendings.remove(query_key)
        except KeyError:
            pass

        # as one task has been released, we can perform allocation again
        self._allocator_ref.enable_quota(_tell=True)
        self._allocator_ref.allocate_tasks(_tell=True)

    def get_allocated_count(self):
        """
        Get total number of operands allocated to run and already running
        """
        return len(self._allocated) + len(self._allocate_pendings)

    def pop_next_request(self):
        """
        Get next unscheduled item from queue. If nothing found, None will
        be returned
        """
        item = None
        while self._req_heap:
            item = heapq.heappop(self._req_heap)
            query_key = (item.session_id, item.op_key)
            # if item is already scheduled or removed, we find next
            if query_key in self._requests:
                del self._requests[query_key]
                break
        return item


class TaskQueueAllocatorActor(WorkerActor):
    """
    Actor performing periodical assignment
    """
    def __init__(self, queue_ref, parallel_num):
        super(TaskQueueAllocatorActor, self).__init__()
        self._parallel_num = parallel_num
        self._has_quota = True

        self._queue_ref = queue_ref
        self._mem_quota_ref = None
        self._execution_ref = None
        self._last_memory_available = 0
        self._last_allocate_time = time.time() - 2

    def post_create(self):
        super(TaskQueueAllocatorActor, self).post_create()

        from .quota import MemQuotaActor
        from .execution import ExecutionActor

        self._queue_ref = self.ctx.actor_ref(self._queue_ref)
        self._mem_quota_ref = self.promise_ref(MemQuotaActor.default_name())
        self._execution_ref = self.ctx.actor_ref(ExecutionActor.default_name())

        self.ref().allocate_tasks(periodical=True, _delay=_ALLOCATE_PERIOD, _tell=True)

    def enable_quota(self):
        self._has_quota = True

    @log_unhandled
    def allocate_tasks(self, periodical=False):
        # make sure the allocation period is not too dense
        if periodical and self._last_allocate_time > time.time() - _ALLOCATE_PERIOD:
            return
        cur_mem_available = resource.virtual_memory().available
        if cur_mem_available > self._last_memory_available:
            # memory usage reduced: try reallocate existing requests
            self._has_quota = True
        self._last_memory_available = cur_mem_available

        num_cpu = resource.cpu_count()
        cpu_rate = resource.cpu_percent()
        batch_allocated = 0
        while self._has_quota:
            allocated_count = self._queue_ref.get_allocated_count()
            if allocated_count >= self._parallel_num:
                break
            if allocated_count >= num_cpu / 4 and num_cpu * 100 - 50 < cpu_rate + batch_allocated:
                break
            item = self._queue_ref.pop_next_request()
            if item is None:
                break

            # obtain quota sizes for operands
            quota_request = self._execution_ref.prepare_quota_request(item.session_id, item.op_key)
            logger.debug('Quota request for %s: %r', item.op_key, quota_request)
            if quota_request:
                local_cb = ((self._queue_ref.uid, self._queue_ref.address),
                            TaskQueueActor.handle_allocated.__name__,
                            item.session_id, item.op_key, item.callback)
                self._queue_ref.mark_allocate_pending(item.session_id, item.op_key)
                self._has_quota = self._mem_quota_ref.request_batch_quota(quota_request, local_cb)
                batch_allocated += 1
            elif quota_request is None:
                # already processed, we skip to the next
                self.ctx.sleep(0.001)
                continue
            else:
                # allocate directly when no quota needed
                self._queue_ref.handle_allocated(item.session_id, item.op_key, item.callback, _tell=True)
                batch_allocated += 1
            self.ctx.sleep(0.001)

        self._last_allocate_time = time.time()
        self.ref().allocate_tasks(periodical=True, _delay=_ALLOCATE_PERIOD, _tell=True)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core import TilesableOperandMixin
from ..core import DataFrameChunkData, DataFrameChunk, DataFrameData, DataFrame


class DataFrameOperandMixin(TilesableOperandMixin):
    __slots__ = ()
    _op_module_ = 'dataframe'

    def _create_chunk(self, output_idx, index, shape, **kw):
        data = DataFrameChunkData(_index=index, _shape=shape, _op=self,
                                  _dtypes=kw.pop('dtypes', None),
                                  _index_value=kw.pop('index_value', None), **kw)
        return DataFrameChunk(data)

    def _create_entity(self, output_idx, shape, nsplits, chunks, **kw):
        if nsplits is not None:
            kw['_nsplits'] = nsplits
        data = DataFrameData(_shape=shape, _op=self, _chunks=chunks,
                             _dtypes=kw.pop('dtypes', None),
                             _index_value=kw.pop('index_value', None), **kw)
        return DataFrame(data)

    def new_dataframes(self, inputs, shape, dtypes=None, index_value=None, columns=None,
                       chunks=None, nsplits=None, output_limit=None, kws=None, **kw):
        return self.new_entities(inputs, shape, dtypes=dtypes, index_value=index_value, columns=columns,
                                 chunks=chunks, nsplits=nsplits,
                                 output_limit=output_limit, kws=kws, **kw)

    def new_dataframe(self, inputs, shape, dtypes=None, index_value=None, columns=None, **kw):
        if getattr(self, 'output_limit') != 1:
            raise TypeError('cannot new tensor with more than 1 outputs')

        return self.new_dataframes(inputs, shape, dtypes=dtypes,
                                   index_value=index_value, columns=columns, **kw)[0]



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import logging
import random
import sys
import time
from functools import partial
from collections import defaultdict

from .. import promise
from ..compat import Enum
from ..config import options
from ..errors import PinChunkFailed, WorkerProcessStopped, ExecutionInterrupted, DependencyMissing
from ..tensor.expressions.datasource import TensorFetchChunk
from ..utils import deserialize_graph, log_unhandled
from .chunkholder import ensure_chunk
from .spill import spill_exists
from .utils import WorkerActor, ExpiringCache, concat_operand_keys

logger = logging.getLogger(__name__)


class ExecutionState(Enum):
    ALLOCATING = 'allocating'
    PREPARING_INPUTS = 'preparing_inputs'
    CALCULATING = 'calculating'
    STORING = 'storing'


class GraphExecutionRecord(object):
    """
    Execution records of the graph
    """
    __slots__ = ('graph', 'graph_serialized', '_state', 'op_string', 'targets', 'io_meta',
                 'priority_data', 'data_sizes', 'chunks_use_once', 'state_time',
                 'mem_request', 'pin_request', 'est_finish_time', 'calc_actor_uid',
                 'send_addresses', 'retry_delay', 'enqueue_callback', 'finish_callbacks',
                 'stop_requested')

    def __init__(self, graph_serialized, state, targets=None, io_meta=None, priority_data=None,
                 data_sizes=None, chunks_use_once=None, mem_request=None, pin_request=None,
                 est_finish_time=None, calc_actor_uid=None, send_addresses=None,
                 retry_delay=None, enqueue_callback=None, finish_callbacks=None,
                 stop_requested=False):
        self.graph_serialized = graph_serialized
        graph = self.graph = deserialize_graph(graph_serialized)

        self._state = state
        self.state_time = time.time()
        self.targets = targets or []
        self.io_meta = io_meta or dict()
        self.data_sizes = data_sizes or dict()
        self.priority_data = priority_data or dict()
        self.chunks_use_once = chunks_use_once or set()
        self.mem_request = mem_request or dict()
        self.pin_request = pin_request or set()
        self.est_finish_time = est_finish_time or time.time()
        self.calc_actor_uid = calc_actor_uid
        self.send_addresses = send_addresses
        self.retry_delay = retry_delay or 0
        self.enqueue_callback = enqueue_callback
        self.finish_callbacks = finish_callbacks or []
        self.stop_requested = stop_requested or False

        _, self.op_string = concat_operand_keys(graph)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value
        self.state_time = time.time()


class GraphResultRecord(object):
    """
    Execution result of a graph
    """
    __slots__ = 'data_sizes', 'exc', 'accept'

    def __init__(self, *args, **kwargs):
        accept = self.accept = kwargs.pop('_accept', True)
        if accept:
            self.data_sizes = args[0]
        else:
            self.exc = args

    def build_args(self):
        if self.accept:
            return (self.data_sizes,), {}
        else:
            return self.exc, dict(_accept=False)


class ExecutionActor(WorkerActor):
    """
    Actor for execution control
    """
    _last_dump_time = time.time()

    def __init__(self):
        super(ExecutionActor, self).__init__()
        self._chunk_holder_ref = None
        self._dispatch_ref = None
        self._task_queue_ref = None
        self._mem_quota_ref = None
        self._status_ref = None
        self._daemon_ref = None

        self._graph_records = dict()  # type: dict[tuple, GraphExecutionRecord]
        self._result_cache = ExpiringCache()  # type: dict[tuple, GraphResultRecord]

    def post_create(self):
        from .chunkholder import ChunkHolderActor
        from .daemon import WorkerDaemonActor
        from .dispatcher import DispatchActor
        from .quota import MemQuotaActor
        from .status import StatusActor
        from .taskqueue import TaskQueueActor

        super(ExecutionActor, self).post_create()
        self._chunk_holder_ref = self.promise_ref(ChunkHolderActor.default_name())
        self._dispatch_ref = self.promise_ref(DispatchActor.default_name())
        self._task_queue_ref = self.promise_ref(TaskQueueActor.default_name())
        self._mem_quota_ref = self.promise_ref(MemQuotaActor.default_name())

        self._daemon_ref = self.ctx.actor_ref(WorkerDaemonActor.default_name())
        if self.ctx.has_actor(self._daemon_ref):
            self._daemon_ref.register_callback(self.ref(), self.handle_process_down.__name__, _tell=True)
        else:
            self._daemon_ref = None

        self._status_ref = self.ctx.actor_ref(StatusActor.default_name())
        if not self.ctx.has_actor(self._status_ref):
            self._status_ref = None

        self.periodical_dump()

    def periodical_dump(self):
        """
        Periodically dump debug information
        """
        if logger.getEffectiveLevel() > logging.DEBUG:
            return
        cls = type(self)
        if cls._last_dump_time < time.time() - 10:
            cls._last_dump_time = time.time()
            if self._graph_records:
                self._dump_execution_states()
        self.ref().periodical_dump(_tell=True, _delay=10)

    @promise.reject_on_exception
    @log_unhandled
    def enqueue_graph(self, session_id, graph_key, graph_ser, io_meta, data_sizes,
                      priority_data=None, send_addresses=None, callback=None):
        """
        Submit graph to the worker and control the execution
        :param session_id: session id
        :param graph_key: graph key
        :param graph_ser: serialized executable graph
        :param io_meta: io meta of the chunk
        :param data_sizes: data size of each input chunk, as a dict
        :param priority_data: data priority
        :param send_addresses: targets to send results after execution
        :param callback: promise callback
        """
        priority_data = priority_data or dict()

        graph_record = self._graph_records[(session_id, graph_key)] = GraphExecutionRecord(
            graph_ser, ExecutionState.ALLOCATING,
            io_meta=io_meta,
            data_sizes=data_sizes,
            enqueue_callback=callback,
            priority_data=priority_data,
            targets=io_meta['chunks'],
            chunks_use_once=set(io_meta.get('input_chunks', [])) - set(io_meta.get('shared_input_chunks', [])),
            send_addresses=send_addresses,
        )
        logger.debug('Worker graph %s(%s) targeting at %r accepted.', graph_key,
                     graph_record.op_string, graph_record.targets)

        self._update_state(session_id, graph_key, ExecutionState.ALLOCATING)

        self._task_queue_ref.enqueue_task(session_id, graph_key, priority_data, _promise=True) \
            .then(lambda *_: self.tell_promise(callback) if callback else None)

    @log_unhandled
    def prepare_quota_request(self, session_id, graph_key):
        """
        Calculate quota request for an execution graph
        :param session_id: session id
        :param graph_key: key of the execution graph
        :return: allocation dict
        """
        try:
            graph_record = self._graph_records[(session_id, graph_key)]
        except KeyError:
            return None

        graph = graph_record.graph
        alloc_mem_batch = dict()
        alloc_cache_batch = dict()
        input_chunk_keys = dict()

        if self._status_ref:
            self.estimate_graph_finish_time(session_id, graph_key)

        # collect potential allocation sizes
        for chunk in graph:
            if not isinstance(chunk.op, TensorFetchChunk) and chunk.key in graph_record.targets:
                # use estimated size as potential allocation size
                alloc_mem_batch[chunk.key] = chunk.rough_nbytes * 2
                alloc_cache_batch[chunk.key] = chunk.rough_nbytes
            else:
                # use actual size as potential allocation size
                input_chunk_keys[chunk.key] = graph_record.data_sizes.get(chunk.key, chunk.nbytes)

        keys_to_pin = list(input_chunk_keys.keys())
        try:
            graph_record.pin_request = set(self._chunk_holder_ref.pin_chunks(graph_key, keys_to_pin))
        except PinChunkFailed:
            # cannot pin input chunks: retry later
            self.dequeue_graph(session_id, graph_key)

            retry_delay = graph_record.retry_delay + 0.5 + random.random()
            graph_record.retry_delay = min(1 + graph_record.retry_delay, 30)
            self.ref().enqueue_graph(
                session_id, graph_key, graph_record.graph_serialized, graph_record.io_meta,
                graph_record.data_sizes, priority_data=graph_record.priority_data,
                send_addresses=graph_record.send_addresses, callback=graph_record.enqueue_callback,
                _tell=True, _delay=retry_delay)
            return None

        load_chunk_sizes = dict((k, v) for k, v in input_chunk_keys.items()
                                if k not in graph_record.pin_request)
        alloc_mem_batch.update((self._build_load_key(graph_key, k), v)
                               for k, v in load_chunk_sizes.items() if k in graph_record.chunks_use_once)
        self._chunk_holder_ref.spill_size(sum(alloc_cache_batch.values()), _tell=True)

        if alloc_mem_batch:
            graph_record.mem_request = alloc_mem_batch
        return alloc_mem_batch

    @log_unhandled
    def dequeue_graph(self, session_id, graph_key):
        """
        Remove execution graph task from queue
        :param session_id: session id
        :param graph_key: key of the execution graph
        """
        self._cleanup_graph(session_id, graph_key)

    @log_unhandled
    def update_priority(self, session_id, graph_key, priority_data):
        """
        Update priority data for given execution graph
        :param session_id: session id
        :param graph_key: key of the execution graph
        :param priority_data: priority data
        """
        query_key = (session_id, graph_key)
        if query_key not in self._graph_records:
            return
        self._graph_records[query_key].priority_data = priority_data
        self._task_queue_ref.update_priority(session_id, graph_key, priority_data)

    @staticmethod
    def _build_load_key(graph_key, chunk_key):
        return '%s_load_memory_%s' % (graph_key, chunk_key)

    @log_unhandled
    def _fetch_remote_data(self, session_id, graph_key, chunk_key, remote_addr, *_, **kwargs):
        """
        Asynchronously send data receiving command to a remote address
        :param session_id: session id
        :param graph_key: graph key
        :param chunk_key: chunk key
        :param remote_addr: remote server containing provided chunk key
        :return: promise object
        """
        from .dispatcher import DispatchActor

        remote_disp_ref = self.promise_ref(uid=DispatchActor.default_name(),
                                           address=remote_addr)
        ensure_cached = kwargs.pop('ensure_cached', True)

        @log_unhandled
        def _finish_fetch(*_):
            self._chunk_holder_ref.pin_chunks(graph_key, chunk_key)
            if self._chunk_holder_ref.is_stored(chunk_key):
                self._mem_quota_ref.release_quota(self._build_load_key(graph_key, chunk_key))

        @log_unhandled
        def _fetch_step(sender_uid):
            if self._graph_records[(session_id, graph_key)].stop_requested:
                self._dispatch_ref.register_free_slot(sender_uid, 'sender')
                raise ExecutionInterrupted

            sender_ref = self.promise_ref(sender_uid, address=remote_addr)
            logger.debug('Request for chunk %s transferring from %s', chunk_key, remote_addr)
            return sender_ref.send_data(
                session_id, chunk_key, self.address, ensure_cached=ensure_cached,
                timeout=options.worker.prepare_data_timeout, _promise=True
            ).then(_finish_fetch)

        return promise.Promise(done=True) \
            .then(lambda *_: remote_disp_ref.get_free_slot('sender', _promise=True)) \
            .then(_fetch_step)

    def estimate_graph_finish_time(self, session_id, graph_key, calc_fetch=True, base_time=None):
        """
        Calc predictions for given chunk graph
        """
        session_graph_key = (session_id, graph_key)
        if session_graph_key not in self._graph_records:
            return
        graph_record = self._graph_records[session_graph_key]
        graph = graph_record.graph

        ops = set(type(c.op).__name__ for c in graph if not isinstance(c.op, TensorFetchChunk))
        op_calc_key = ('calc_speed.' + list(ops)[0]) if len(ops) == 1 else None

        stats = defaultdict(lambda: dict(count=0))
        if self._status_ref:
            stats.update(self._status_ref.get_stats(['disk_read_speed', 'disk_write_speed',
                                                     'net_transfer_speed', op_calc_key]))

        if op_calc_key not in stats:
            return None
        if stats[op_calc_key]['count'] < options.optimize.min_stats_count:
            return None
        if abs(stats[op_calc_key]['count']) < 1e-6:
            return None

        input_size = 0
        net_size = 0
        disk_size = 0
        base_time = base_time or time.time()

        if calc_fetch:
            for c in graph:
                if not isinstance(c.op, TensorFetchChunk):
                    break
                input_size += c.nbytes
                if self._chunk_holder_ref.is_stored(c.key):
                    continue
                if spill_exists(c.key):
                    disk_size += c.nbytes
                else:
                    net_size += c.nbytes

            if stats['net_transfer_speed']['count'] >= options.optimize.min_stats_count:
                base_time += net_size * 1.0 / stats['net_transfer_speed']['mean']
            if stats['disk_read_speed']['count'] >= options.optimize.min_stats_count:
                base_time += disk_size * 1.0 / stats['disk_read_speed']['mean']
            else:
                base_time += disk_size * 1.0 / options.optimize.default_disk_io_speed

        est_finish_time = base_time + input_size * 1.0 / stats[op_calc_key]['mean']

        graph_record.est_finish_time = est_finish_time
        self._status_ref.update_stats(dict(
            min_est_finish_time=min(rec.est_finish_time for rec in self._graph_records.values()),
            max_est_finish_time=max(rec.est_finish_time for rec in self._graph_records.values()),
        ), _tell=True, _wait=False)

        self.ref().estimate_graph_finish_time(session_id, graph_key, _tell=True, _delay=1)

    def _update_state(self, session_id, key, state):
        logger.debug('Operand %s switched to %s', key, getattr(state, 'name'))
        record = self._graph_records[(session_id, key)]
        record.state = state
        if self._status_ref:
            self._status_ref.update_progress(session_id, key, record.op_string, state.name,
                                             _tell=True, _wait=False)

    @promise.reject_on_exception
    @log_unhandled
    def start_execution(self, session_id, graph_key, send_addresses=None, callback=None):
        """
        Submit graph to the worker and control the execution
        :param session_id: session id
        :param graph_key: key of the execution graph
        :param send_addresses: targets to send results after execution
        :param callback: promise callback
        """
        graph_record = self._graph_records[(session_id, graph_key)]
        if send_addresses:
            graph_record.send_addresses = send_addresses

        # add callbacks to callback store
        if callback is None:
            callback = []
        elif not isinstance(callback, list):
            callback = [callback]
        graph_record.finish_callbacks.extend(callback)
        try:
            del self._result_cache[(session_id, graph_key)]
        except KeyError:
            pass

        @log_unhandled
        def _wait_free_slot(*_):
            return self._dispatch_ref.get_free_slot('cpu', _promise=True)

        @log_unhandled
        def _handle_rejection(*exc):
            # some error occurred...
            logger.debug('Entering _handle_rejection() for graph %s', graph_key)
            self._dump_execution_states()

            if graph_record.stop_requested:
                graph_record.stop_requested = False
                if not isinstance(exc[1], ExecutionInterrupted):
                    try:
                        raise ExecutionInterrupted
                    except ExecutionInterrupted:
                        exc = sys.exc_info()

            if isinstance(exc[1], ExecutionInterrupted):
                logger.warning('Execution of graph %s interrupted.', graph_key)
            else:
                logger.exception('Unexpected error occurred in executing %s', graph_key, exc_info=exc)

            self._result_cache[(session_id, graph_key)] = GraphResultRecord(*exc, **dict(_accept=False))
            self._invoke_finish_callbacks(session_id, graph_key)

        self._prepare_graph_inputs(session_id, graph_key) \
            .then(_wait_free_slot) \
            .then(lambda uid: self._send_calc_request(session_id, graph_key, uid)) \
            .then(lambda uid, sizes: self._dump_cache(session_id, graph_key, uid, sizes)) \
            .then(lambda *_: self._invoke_finish_callbacks(session_id, graph_key)) \
            .catch(_handle_rejection)

    @log_unhandled
    def _prepare_graph_inputs(self, session_id, graph_key):
        """
        Load input data from spilled storage and other workers
        :param session_id: session id
        :param graph_key: key of the execution graph
        """
        graph_record = self._graph_records[(session_id, graph_key)]
        if graph_record.stop_requested:
            raise ExecutionInterrupted

        unspill_keys = []
        transfer_keys = []

        logger.debug('Start preparing input data for graph %s', graph_key)
        self._update_state(session_id, graph_key, ExecutionState.PREPARING_INPUTS)
        prepare_promises = []
        chunks_use_once = graph_record.chunks_use_once

        handled_keys = set()
        for chunk in graph_record.graph:
            if not isinstance(chunk.op, TensorFetchChunk):
                continue
            if chunk.key in handled_keys:
                continue
            handled_keys.add(chunk.key)

            if self._chunk_holder_ref.is_stored(chunk.key):
                # data already in plasma: we just pin it
                pinned_keys = self._chunk_holder_ref.pin_chunks(graph_key, chunk.key)
                if chunk.key in pinned_keys:
                    self._mem_quota_ref.release_quota(self._build_load_key(graph_key, chunk.key))
                    continue

            if spill_exists(chunk.key):
                if chunk.key in chunks_use_once:
                    # input only use in current operand, we only need to load it into process memory
                    continue
                self._mem_quota_ref.release_quota(self._build_load_key(graph_key, chunk.key))
                load_fun = partial(lambda gk, ck, *_: self._chunk_holder_ref.pin_chunks(gk, ck),
                                   graph_key, chunk.key)
                unspill_keys.append(chunk.key)
                prepare_promises.append(ensure_chunk(self, session_id, chunk.key, move_to_end=True) \
                                        .then(load_fun))
                continue

            # load data from another worker
            chunk_meta = self.get_meta_ref(session_id, chunk.key) \
                .get_chunk_meta(session_id, chunk.key)
            if chunk_meta is None:
                raise DependencyMissing('Dependency %s not met on sending.' % chunk.key)
            worker_results = chunk_meta.workers

            worker_priorities = []
            for worker_ip in worker_results:
                # todo sort workers by speed of network and other possible factors
                worker_priorities.append((worker_ip, (0, )))

            transfer_keys.append(chunk.key)

            # fetch data from other workers, if one fails, try another
            sorted_workers = sorted(worker_priorities, key=lambda pr: pr[1])
            p = self._fetch_remote_data(session_id, graph_key, chunk.key, sorted_workers[0][0],
                                        ensure_cached=chunk.key not in chunks_use_once)
            for wp in sorted_workers[1:]:
                p = p.catch(functools.partial(self._fetch_remote_data, session_id, graph_key, chunk.key, wp[0],
                                              ensure_cached=chunk.key not in chunks_use_once))
            prepare_promises.append(p)

        logger.debug('Graph key %s: Targets %r, unspill keys %r, transfer keys %r',
                     graph_key, graph_record.targets, unspill_keys, transfer_keys)
        return promise.all_(prepare_promises)

    @log_unhandled
    def _send_calc_request(self, session_id, graph_key, calc_uid):
        """
        Start actual calculation in CpuCalcActor
        :param session_id: session id
        :param graph_key: key of the execution graph
        :param calc_uid: uid of the allocated CpuCalcActor
        """
        graph_record = self._graph_records[(session_id, graph_key)]
        try:
            if graph_record.stop_requested:
                raise ExecutionInterrupted

            graph_record.calc_actor_uid = calc_uid

            # get allocation for calc, in case that memory exhausts
            target_allocs = dict()
            for chunk in graph_record.graph:
                if isinstance(chunk.op, TensorFetchChunk):
                    if not self._chunk_holder_ref.is_stored(chunk.key):
                        alloc_key = self._build_load_key(graph_key, chunk.key)
                        if alloc_key in graph_record.mem_request:
                            target_allocs[alloc_key] = graph_record.mem_request[alloc_key]
                elif chunk.key in graph_record.targets:
                    target_allocs[chunk.key] = graph_record.mem_request[chunk.key]

            logger.debug('Start calculation for graph %s in actor %s', graph_key, calc_uid)

            self._update_state(session_id, graph_key, ExecutionState.CALCULATING)
            raw_calc_ref = self.ctx.actor_ref(calc_uid)
            calc_ref = self.promise_ref(raw_calc_ref)

            def _start_calc(*_):
                if self._daemon_ref is None or self._daemon_ref.is_actor_process_alive(raw_calc_ref):
                    return calc_ref.calc(session_id, graph_record.graph_serialized,
                                         graph_record.targets, _promise=True)
                else:
                    raise WorkerProcessStopped

            self.estimate_graph_finish_time(session_id, graph_key, calc_fetch=False)
        except:
            self._dispatch_ref.register_free_slot(calc_uid, 'cpu')
            raise

        # make sure that memory suffices before actually run execution
        return self._mem_quota_ref.request_batch_quota(target_allocs, _promise=True) \
            .then(_start_calc)

    @log_unhandled
    def _dump_cache(self, session_id, graph_key, inproc_uid, save_sizes):
        """
        Dump calc results into shared cache or spill
        :param session_id: session id
        :param graph_key: key of the execution graph
        :param inproc_uid: uid of the InProcessCacheActor
        :param save_sizes: sizes of data
        """
        graph_record = self._graph_records[session_id, graph_key]
        calc_keys = graph_record.targets
        send_addresses = graph_record.send_addresses

        @log_unhandled
        def _do_active_transfer(*_):
            # transfer the result chunk to expected endpoints
            @log_unhandled
            def _send_chunk(sender_uid, chunk_key, target_addrs):
                if graph_record.stop_requested:
                    self._dispatch_ref.register_free_slot(sender_uid, 'sender')
                    raise ExecutionInterrupted

                sender_ref = self.promise_ref(sender_uid)
                logger.debug('Request for chunk %s sent to %s', chunk_key, target_addrs)
                return sender_ref.send_data(session_id, chunk_key, target_addrs, ensure_cached=False,
                                            timeout=options.worker.prepare_data_timeout, _promise=True)

            if graph_record.mem_request:
                self._mem_quota_ref.release_quotas(tuple(graph_record.mem_request.keys()), _tell=True)
                graph_record.mem_request = dict()

            promises = []
            for key, targets in send_addresses.items():
                promises.append(self._dispatch_ref.get_free_slot('sender', _promise=True)
                                .then(partial(_send_chunk, chunk_key=key, target_addrs=targets))
                                .catch(lambda *_: None))
            return promise.all_(promises)

        logger.debug('Graph %s: Start putting %r into shared cache. Target actor uid %s.',
                     graph_key, calc_keys, inproc_uid)
        self._update_state(session_id, graph_key, ExecutionState.STORING)

        raw_inproc_ref = self.ctx.actor_ref(inproc_uid)
        inproc_ref = self.promise_ref(raw_inproc_ref)

        if graph_record.stop_requested:
            logger.debug('Graph %s already marked for stop, quit.', graph_key)
            if (self._daemon_ref is None or self._daemon_ref.is_actor_process_alive(raw_inproc_ref)) \
                    and self.ctx.has_actor(raw_inproc_ref):
                logger.debug('Try remove keys for graph %s.', graph_key)
                raw_inproc_ref.remove_cache(list(calc_keys), _tell=True)
            logger.debug('Graph %s already marked for stop, quit.', graph_key)
            raise ExecutionInterrupted

        self._chunk_holder_ref.unpin_chunks(
            graph_key, list(set(c.key for c in graph_record.graph)), _tell=True)
        self._dump_execution_states()

        if self._daemon_ref is not None and not self._daemon_ref.is_actor_process_alive(raw_inproc_ref):
            raise WorkerProcessStopped

        def _cache_result(*_):
            self._result_cache[(session_id, graph_key)] = GraphResultRecord(save_sizes)

        if not send_addresses:
            # no endpoints to send, dump keys into shared memory and return
            logger.debug('Worker graph %s(%s) finished execution. Dumping %r into plasma...',
                         graph_key, graph_record.op_string, calc_keys)
            return inproc_ref.dump_cache(calc_keys, _promise=True) \
                .then(_cache_result)
        else:
            # dump keys into shared memory and send
            logger.debug('Worker graph %s(%s) finished execution. Dumping %r into plasma '
                         'while actively transferring %r...',
                         graph_key, graph_record.op_string, calc_keys, send_addresses)

            return inproc_ref.dump_cache(calc_keys, _promise=True) \
                .then(_do_active_transfer) \
                .then(_cache_result)

    def _cleanup_graph(self, session_id, graph_key):
        """
        Do clean up after graph is executed
        :param session_id: session id
        :param graph_key: graph key
        """
        logger.debug('Cleaning callbacks for graph %s', graph_key)
        self._task_queue_ref.release_task(session_id, graph_key, _tell=True)

        try:
            graph_record = self._graph_records[(session_id, graph_key)]
        except KeyError:
            return

        self._mem_quota_ref.cancel_requests(tuple(graph_record.mem_request.keys()), _tell=True)
        if graph_record.mem_request:
            self._mem_quota_ref.release_quotas(tuple(graph_record.mem_request.keys()), _tell=True)
        if graph_record.pin_request:
            self._chunk_holder_ref.unpin_chunks(graph_key, graph_record.pin_request, _tell=True)

        if self._status_ref:
            self._status_ref.remove_progress(session_id, graph_key, _tell=True, _wait=False)
        del self._graph_records[(session_id, graph_key)]

    @promise.reject_on_exception
    @log_unhandled
    def add_finish_callback(self, session_id, graph_key, callback):
        """
        Register a callback to callback store
        :param session_id: session id
        :param graph_key: graph key
        :param callback: promise call
        """
        logger.debug('Adding callback %r for graph %s', callback, graph_key)
        try:
            args, kwargs = self._result_cache[(session_id, graph_key)].build_args()
            self.tell_promise(callback, *args, **kwargs)
        except KeyError:
            self._graph_records[(session_id, graph_key)].finish_callbacks.append(callback)

    @log_unhandled
    def stop_execution(self, session_id, graph_key):
        """
        Mark graph for stopping
        :param graph_key: graph key
        """
        logger.debug('Receive stop for graph %s', graph_key)
        try:
            graph_record = self._graph_records[(session_id, graph_key)]
        except KeyError:
            return

        graph_record.stop_requested = True
        if graph_record.state == ExecutionState.ALLOCATING:
            try:
                raise ExecutionInterrupted
            except:
                exc_info = sys.exc_info()
            if graph_record.mem_request:
                self._mem_quota_ref.cancel_requests(tuple(graph_record.mem_request.keys()), exc_info, _tell=True)
        elif graph_record.state == ExecutionState.CALCULATING:
            if self._daemon_ref is not None and graph_record.calc_actor_uid is not None:
                self._daemon_ref.kill_actor_process(self.ctx.actor_ref(graph_record.calc_actor_uid), _tell=True)

    @log_unhandled
    def _invoke_finish_callbacks(self, session_id, graph_key):
        """
        Call finish callback when execution is done
        :param session_id: session id
        :param graph_key: graph key
        """
        query_key = (session_id, graph_key)
        callbacks = self._graph_records[query_key].finish_callbacks
        args, kwargs = self._result_cache[query_key].build_args()
        logger.debug('Send finish callback for graph %s into %d targets', graph_key, len(callbacks))
        for cb in callbacks:
            self.tell_promise(cb, *args, **kwargs)
        self._cleanup_graph(session_id, graph_key)

    def _dump_execution_states(self):
        if logger.getEffectiveLevel() <= logging.DEBUG:
            cur_time = time.time()
            states = dict((k[1], (cur_time - v.state_time, v.state.name))
                          for k, v in self._graph_records.items()
                          if v.state != ExecutionState.ALLOCATING)
            logger.debug('Executing states: %r', states)

    def handle_process_down(self, halt_refs):
        """
        Handle process down event
        :param halt_refs: actor refs in halt processes
        """
        logger.debug('Process halt detected. Trying to reject affected promises %r.',
                     [ref.uid for ref in halt_refs])
        try:
            raise WorkerProcessStopped
        except WorkerProcessStopped:
            exc_info = sys.exc_info()

        for ref in halt_refs:
            self.reject_promise_ref(ref, *exc_info)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import math
import os
import time
from collections import OrderedDict

from ..actors import ActorNotExist
from ..compat import OrderedDict3
from ..config import options
from ..promise import PromiseActor
from ..cluster_info import HasClusterInfoActor


logger = logging.getLogger(__name__)


class WorkerActor(HasClusterInfoActor, PromiseActor):
    """
    Base class of all worker actors, providing necessary utils
    """
    @classmethod
    def default_name(cls):
        return 'w:{0}'.format(cls.__name__)

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())
        try:
            self.set_cluster_info_ref()
        except ActorNotExist:
            pass
        self._init_chunk_store()

    def _init_chunk_store(self):
        import pyarrow.plasma as plasma
        from .chunkstore import PlasmaChunkStore
        self._plasma_client = plasma.connect(options.worker.plasma_socket, '', 0)
        self._chunk_store = PlasmaChunkStore(self._plasma_client)

    def get_meta_ref(self, session_id, chunk_key):
        from ..scheduler.chunkmeta import LocalChunkMetaActor
        addr = self.get_scheduler((session_id, chunk_key))
        return self.ctx.actor_ref(LocalChunkMetaActor.default_name(), address=addr)


class ExpMeanHolder(object):
    """
    Collector of statistics of a given series. The value decays by _factor as time elapses.
    """
    def __init__(self, factor=0.8):
        self._factor = factor
        self._count = 0
        self._v_divided = 0
        self._v_divisor = 0
        self._v2_divided = 0

    def put(self, value):
        self._count += 1
        self._v_divided = self._v_divided * self._factor + value
        self._v_divisor = self._v_divisor * self._factor + 1
        self._v2_divided = self._v2_divided * self._factor + value ** 2

    def count(self):
        return self._count

    def mean(self):
        if self._count == 0:
            return 0
        return self._v_divided * 1.0 / self._v_divisor

    def var(self):
        if self._count == 0:
            return 0
        return self._v2_divided * 1.0 / self._v_divisor - self.mean() ** 2

    def std(self):
        return math.sqrt(self.var())


class ExpiringCache(dict):
    def __init__(self, *args, **kwargs):
        expire_time = kwargs.pop('_expire_time', options.worker.callback_preserve_time)
        super(ExpiringCache, self).__init__(*args, **kwargs)

        self._expire_time = expire_time
        self._insert_times = OrderedDict3()

    def __setitem__(self, key, value):
        super(ExpiringCache, self).__setitem__(key, value)
        if key in self._insert_times:
            self._insert_times[key] = time.time()
            self._insert_times.move_to_end(key)
            return

        clean_keys = []
        self._insert_times[key] = time.time()
        last_finish_time = time.time() - self._expire_time
        for k, t in self._insert_times.items():
            if t < last_finish_time:
                clean_keys.append(k)
            else:
                break
        for k in clean_keys:
            del self[k]


def concat_operand_keys(graph, sep=','):
    from ..tensor.expressions.datasource import TensorFetchChunk
    graph_op_dict = OrderedDict()
    for c in graph:
        if isinstance(c.op, TensorFetchChunk):
            continue
        graph_op_dict[c.op.key] = type(c.op).__name__
    keys = sep.join(graph_op_dict.keys())
    graph_ops = sep.join(graph_op_dict.values())
    return keys, graph_ops
